@@ -1,10 +1,13 @@
 #include <cog.h>
 
 #include <assert.h>
+#include <errno.h>
 #include <fcntl.h>
 #include <gbm.h>
 #include <libinput.h>
 #include <libudev.h>
+#include <string.h>
+#include <wayland-server.h>
 #include <wpe/fdo.h>
 #include <wpe/fdo-egl.h>
 #include <xf86drm.h>
@@ -24,6 +27,9 @@ typedef EGLDisplay (EGLAPIENTRYP PFNEGLGETPLATFORMDISPLAYEXTPROC) (EGLenum platf
 
 
 struct buffer_object {
+    struct wl_list link;
+    struct wl_listener destroy_listener;
+
     uint32_t fb_id;
     struct gbm_bo *bo;
     struct wl_resource *buffer_resource;
@@ -44,6 +50,7 @@ static struct {
     uint32_t height;
 
     bool mode_set;
+    struct wl_list buffer_list;
     struct buffer_object *committed_buffer;
 } drm_data = {
     .fd = -1,
@@ -107,7 +114,8 @@ static struct {
 } wpe_view_data;
 
 
-static void destroy_buffer (struct buffer_object *buffer)
+static void
+destroy_buffer (struct buffer_object *buffer)
 {
     drmModeRmFB (drm_data.fd, buffer->fb_id);
     gbm_bo_destroy (buffer->bo);
@@ -116,9 +124,32 @@ static void destroy_buffer (struct buffer_object *buffer)
     g_free (buffer);
 }
 
+static void
+destroy_buffer_notify (struct wl_listener *listener, void *data)
+{
+    struct buffer_object *buffer = wl_container_of (listener, buffer, destroy_listener);
+
+    if (drm_data.committed_buffer == buffer)
+        drm_data.committed_buffer = NULL;
+
+    wl_list_remove (&buffer->link);
+    destroy_buffer (buffer);
+}
+
 static void
 clear_drm (void)
 {
+    drm_data.committed_buffer = NULL;
+
+    struct buffer_object *buffer, *tmp;
+    wl_list_for_each_safe (buffer, tmp, &drm_data.buffer_list, link) {
+        wl_list_remove (&buffer->link);
+        wl_list_remove (&buffer->destroy_listener.link);
+
+        destroy_buffer (buffer);
+    }
+    wl_list_init (&drm_data.buffer_list);
+
     g_clear_pointer (&drm_data.encoder, drmModeFreeEncoder);
     g_clear_pointer (&drm_data.connector, drmModeFreeConnector);
     if (drm_data.fd != -1) {
@@ -213,6 +244,7 @@ init_drm (void)
 
     drm_data.width = drm_data.mode->hdisplay;
     drm_data.height = drm_data.mode->vdisplay;
+    wl_list_init (&drm_data.buffer_list);
 
     g_clear_pointer (&resources, drmModeFreeResources);
     return TRUE;
@@ -221,10 +253,99 @@ init_drm (void)
 static void
 drm_page_flip_handler (int fd, unsigned int frame, unsigned int sec, unsigned int usec, void *data)
 {
-    g_clear_pointer (&drm_data.committed_buffer, destroy_buffer);
+    if (drm_data.committed_buffer)
+        wpe_view_backend_exportable_fdo_dispatch_release_buffer (wpe_host_data.exportable, drm_data.committed_buffer->buffer_resource);
+    drm_data.committed_buffer = (struct buffer_object *) data;
 
     wpe_view_backend_exportable_fdo_dispatch_frame_complete (wpe_host_data.exportable);
-    drm_data.committed_buffer = (struct buffer_object *) data;
+}
+
+static struct buffer_object *
+drm_buffer_for_resource (struct wl_resource *buffer_resource)
+{
+    struct buffer_object *buffer;
+    wl_list_for_each (buffer, &drm_data.buffer_list, link) {
+        if (buffer->buffer_resource == buffer_resource)
+            return buffer;
+    }
+
+    return NULL;
+}
+
+static struct buffer_object *
+drm_create_buffer_for_bo (struct gbm_bo *bo, struct wl_resource *buffer_resource, uint32_t width, uint32_t height, uint32_t format)
+{
+    uint32_t in_handles[4] = { 0, };
+    uint32_t in_strides[4] = { 0, };
+    uint32_t in_offsets[4] = { 0, };
+    uint64_t in_modifiers[4] = { 0, };
+
+    in_modifiers[0] = gbm_bo_get_modifier (bo);
+
+    int plane_count = MIN (gbm_bo_get_plane_count (bo), 4);
+    for (int i = 0; i < plane_count; ++i) {
+        in_handles[i] = gbm_bo_get_handle_for_plane (bo, i).u32;
+        in_strides[i] = gbm_bo_get_stride_for_plane (bo, i);
+        in_offsets[i] = gbm_bo_get_offset (bo, i);
+        in_modifiers[i] = in_modifiers[0];
+    }
+
+    int flags = 0;
+    if (in_modifiers[0])
+        flags = DRM_MODE_FB_MODIFIERS;
+
+    uint32_t fb_id = 0;
+    int ret = drmModeAddFB2WithModifiers (drm_data.fd, width, height, format,
+                                          in_handles, in_strides, in_offsets, in_modifiers,
+                                          &fb_id, flags);
+    if (ret) {
+        in_handles[0] = gbm_bo_get_handle (bo).u32;
+        in_handles[1] = in_handles[2] = in_handles[3] = 0;
+        in_strides[0] = gbm_bo_get_stride (bo);
+        in_strides[1] = in_strides[2] = in_strides[3] = 0;
+        in_offsets[0] = in_offsets[1] = in_offsets[2] = in_offsets[3] = 0;
+
+        ret = drmModeAddFB2 (drm_data.fd, width, height, format,
+                             in_handles, in_strides, in_offsets,
+                             &fb_id, 0);
+    }
+
+    if (ret) {
+        g_warning ("failed to create framebuffer: %s", strerror (errno));
+        return NULL;
+    }
+
+    struct buffer_object *buffer = g_new0 (struct buffer_object, 1);
+    wl_list_insert (&drm_data.buffer_list, &buffer->link);
+    buffer->destroy_listener.notify = destroy_buffer_notify;
+    wl_resource_add_destroy_listener (buffer_resource, &buffer->destroy_listener);
+
+    buffer->fb_id = fb_id;
+    buffer->bo = bo;
+    buffer->buffer_resource = buffer_resource;
+
+    return buffer;
+}
+
+static void
+drm_commit_buffer (struct buffer_object *buffer)
+{
+    int ret;
+    if (!drm_data.mode_set) {
+        ret = drmModeSetCrtc (drm_data.fd, drm_data.crtc_id, buffer->fb_id, 0, 0,
+                              &drm_data.connector_id, 1, drm_data.mode);
+        if (ret) {
+            g_warning ("failed to set mode: %s", strerror (errno));
+            return;
+        }
+
+        drm_data.mode_set = true;
+    }
+
+    ret = drmModePageFlip (drm_data.fd, drm_data.crtc_id, buffer->fb_id,
+                               DRM_MODE_PAGE_FLIP_EVENT, buffer);
+    if (ret)
+        g_warning ("failed to schedule a page flip: %s", strerror (errno));
 }
 
 
@@ -569,12 +690,37 @@ init_glib (void)
 static void
 on_export_buffer_resource (void *data, struct wl_resource *buffer_resource)
 {
-    assert (!"should not be reached");
+    struct buffer_object *buffer = drm_buffer_for_resource (buffer_resource);
+    if (buffer) {
+        drm_commit_buffer (buffer);
+        return;
+    }
+
+    struct gbm_bo* bo = gbm_bo_import (gbm_data.device, GBM_BO_IMPORT_WL_BUFFER,
+                                       (void *) buffer_resource, GBM_BO_USE_SCANOUT);
+    if (!bo) {
+        g_warning ("failed to import a wl_buffer resource into gbm_bo");
+        return;
+    }
+
+    uint32_t width = gbm_bo_get_width (bo);
+    uint32_t height = gbm_bo_get_height (bo);
+    uint32_t format = gbm_bo_get_format (bo);
+
+    buffer = drm_create_buffer_for_bo (bo, buffer_resource, width, height, format);
+    if (buffer)
+        drm_commit_buffer (buffer);
 }
 
 static void
 on_export_dmabuf_resource (void *data, struct wpe_view_backend_exportable_fdo_dmabuf_resource *dmabuf_resource)
 {
+    struct buffer_object *buffer = drm_buffer_for_resource (dmabuf_resource->buffer_resource);
+    if (buffer) {
+        drm_commit_buffer (buffer);
+        return;
+    }
+
     struct gbm_import_fd_modifier_data modifier_data = {
         .width = dmabuf_resource->width,
         .height = dmabuf_resource->height,
@@ -590,55 +736,23 @@ on_export_dmabuf_resource (void *data, struct wpe_view_backend_exportable_fdo_dm
 
     struct gbm_bo *bo = gbm_bo_import (gbm_data.device, GBM_BO_IMPORT_FD_MODIFIER,
                                        (void *)(&modifier_data), GBM_BO_USE_SCANOUT);
-    if (!bo)
+    if (!bo) {
+        g_warning ("failed to import a dma-buf resource into gbm_bo");
         return;
-
-    uint32_t in_handles[4] = { 0, };
-    uint32_t in_strides[4] = { 0, };
-    uint32_t in_offsets[4] = { 0, };
-    uint64_t in_modifiers[4] = { 0, };
-    in_modifiers[0] = gbm_bo_get_modifier (bo);
-
-    int plane_count = gbm_bo_get_plane_count (bo);
-    for (int i = 0; i < plane_count; ++i) {
-        in_handles[i] = gbm_bo_get_handle_for_plane (bo, i).u32;
-        in_strides[i] = gbm_bo_get_stride_for_plane (bo, i);
-        in_offsets[i] = gbm_bo_get_offset (bo, i);
-        in_modifiers[i] = in_modifiers[0];
     }
 
-    int flags = 0;
-    if (in_modifiers[0])
-        flags = DRM_MODE_FB_MODIFIERS;
-
-    uint32_t fb_id = 0;
-    int ret = drmModeAddFB2WithModifiers (drm_data.fd, dmabuf_resource->width, dmabuf_resource->height, dmabuf_resource->format,
-                                          in_handles, in_strides, in_offsets,
-                                          in_modifiers, &fb_id, flags);
-    if (ret)
-        return;
-
-    if (!drm_data.mode_set) {
-        ret = drmModeSetCrtc (drm_data.fd, drm_data.crtc_id, fb_id, 0, 0,
-                              &drm_data.connector_id, 1, drm_data.mode);
-        if (ret)
-            return;
-        drm_data.mode_set = true;
-    }
-
-    struct buffer_object *buffer = g_new0 (struct buffer_object, 1);
-    buffer->fb_id = fb_id;
-    buffer->bo = bo;
-    buffer->buffer_resource = dmabuf_resource->buffer_resource;
-    ret = drmModePageFlip (drm_data.fd, drm_data.crtc_id, fb_id,
-                           DRM_MODE_PAGE_FLIP_EVENT, buffer);
+    buffer = drm_create_buffer_for_bo (bo, dmabuf_resource->buffer_resource,
+                                       dmabuf_resource->width, dmabuf_resource->height,
+                                       dmabuf_resource->format);
+    if (buffer)
+        drm_commit_buffer (buffer);
 }
 
 gboolean
-cog_platform_setup (CogPlatform *platform,
-                    CogShell    *shell G_GNUC_UNUSED,
-                    const char  *params,
-                    GError     **error)
+cog_platform_plugin_setup (CogPlatform *platform,
+                           CogShell    *shell G_GNUC_UNUSED,
+                           const char  *params,
+                           GError     **error)
 {
     g_assert (platform);
     g_return_val_if_fail (COG_IS_SHELL (shell), FALSE);
@@ -697,7 +811,7 @@ cog_platform_setup (CogPlatform *platform,
 }
 
 void
-cog_platform_teardown (CogPlatform *platform)
+cog_platform_plugin_teardown (CogPlatform *platform)
 {
     g_assert (platform);
 
@@ -711,9 +825,9 @@ cog_platform_teardown (CogPlatform *platform)
 }
 
 WebKitWebViewBackend *
-cog_platform_get_view_backend (CogPlatform   *platform,
-                               WebKitWebView *related_view,
-                               GError       **error)
+cog_platform_plugin_get_view_backend (CogPlatform   *platform,
+                                      WebKitWebView *related_view,
+                                      GError       **error)
 {
     static struct wpe_view_backend_exportable_fdo_client exportable_client = {
         .export_buffer_resource = on_export_buffer_resource,
